#ifndef _AUTO_STRING_H_
#define _AUTO_STRING_H_

extern "C" {
	extern	char	*strdup(const char *);
	extern	void	free(void *);
}

/*
   A simple string class which always owns the string it contains.
   The owned string is deleted when the string object is  destroyed.
 */  

class auto_string {
	friend	ostream	&operator<<(ostream &, auto_string &);
	char	*_string;
public:
	auto_string(char *string = 0) {
		_string = strdup(string ? string : "");
	}
	auto_string(char *string, bool given = false) {
		if (given)
			_string = string;
		else
			_string = strdup(string ? string : "");
	}
	~auto_string() {
		free(_string);
		_string = 0;
	}
/*protected: */
	/* give ownership of the char * to the auto_string */
	void	give(char *newer) {
		char	*older = _string;
		_string = newer;
		free(older);
	}

	char	*str() { return _string; }
	void	set(char *string) {
		char	*newer= strdup(string ? string : "");
		give(newer);
	}

	operator const char *() { return _string; }
	char operator[](int n) { return _string[n]; }

	operator =(char *astring) { set(astring); }
};
		
#endif /* _AUTO_STRING_H_ */ 
