#include <stdlib.h>
#include <stdio.h>
#include <ostream.h>

#include "symbol.hh"
#include "types.hh"
#include "typedb.hh"
#include "tree.hh"

#include "oql_context.hh"

extern "C" {
int getopt();
char *optarg;

int	ql_parser(Ql_tree_node **, int);
}

#include "db.hh"


void check_defines(oqlContext &context)
{
	int	token;
	Symbol	*sym;
	Type	*result;
	aqua_t	*define;

	for (token = 0; sym = context.defines.iterate(token); /*void*/) {
#if 0
		// cout << sym->name() << " flags: " << sym->flags() << '\n';
#endif
		if (sym->flags())
			continue;
		define = (aqua_t *) sym->data();
		result = define->typecheck(context);
		cout << "Define " << sym->name() << " : " << *result << '\n';
		cout << *define << '\n';
		sym->setFlags(1);	/* don't typecheck again */
	}
}


main(int argc, char **argv)
{
	int	r;
	int	oql_mode = 1;
	aqua_t	*aqua_tree;
	int	c;
	int	preprint = 0;
	int	convert = 1;
	int	errors = 0;
	int	typecheck = 1;
	char	*_progname = argv[0];
	int	prompting;
	int	hoofer=1;
	int	print_types = 0;
	Type	*query_type;
	oqlContext	context;
	Ql_tree_node	*root;

	while ((c = getopt(argc, argv, "pcPCtTbBvV")) != EOF) {
		switch (c) {
		case 'p':
			preprint++;
			break;
		case 'c':
			convert++;
			break;
		case 't':
			typecheck++;
			break;
		case 'C':
			convert=0;
			break;
		case 'P':
			preprint=0;
			break;
		case 'T':
			typecheck=0;
			break;
		case 'b':
			hoofer=1;
			break;
		case 'B':
			hoofer=0;
			break;
		case 'v':
			print_types = 1;
			break;
		case 'V':
			print_types = 0;
			break;
		default:
			errors++;
			break;
		}
	}
	if (errors) {
		fprintf(stderr,
		"usage: %s  [-p[reprint]] [-c[onvert]] [-t[ypecheck]\n",
			_progname);
		exit(1);
	}

	init_atomic(context.db);
	if (hoofer)
		init_hoofer_db(context.db);
	else
		init_travel_db(context.db);

	aqua_t::setTypePrinting(print_types);
	
	prompting = convert || preprint || typecheck;
	
	if (prompting)
		printf("\n--- Query ----\n");
	r = ql_parser(&root, oql_mode);

	while (r == 0 && root) {
		if (preprint) {
			cout << "\n              --------- Parse Tree ----------\n";
			cout << *root;
		}
		if (convert) {
			aqua_tree = root->aqua_convert(context);
			check_defines(context);
			if (typecheck) {
				query_type = aqua_tree->typecheck(context);
				cout << "Result type: " << *query_type << '\n';
			}
			cout << "\n              ------- Aqua Tree --------\n";
			cout << *aqua_tree;
			cout << '\n'; 
			// delete aqua_tree;
		}
		// delete root;
		root = (Ql_tree_node *)0;

		if (prompting)
			printf("\n--- Query ----\n");
		r = ql_parser(&root, oql_mode);
	}
	exit(0);
}

