#include <stdio.h>

#include "ql_include.h"

static Ql_tree_node *
node_alloc(type, kids)
	enum node_type type;
	int	kids;
{
	Ql_tree_node *node;

	node = (Ql_tree_node *)ql_ma_synt_malloc(sizeof(*node));
	node->type = type;
	node->id = (char *)0;
	node->nkids = kids;
	if (kids)
		node->kids = (Ql_tree_node **)ql_ma_synt_malloc(kids * sizeof(Ql_tree_node *));
	return node;
}

#define	NODE(x, type, kids)	Ql_tree_node *x = node_alloc(type, kids)
						 

Ql_tree_node *
ql_tr_build_node_no_son(type, id)
	int	type;
	char	*id;
{
	NODE(node, type, 0);

	node->id = id;
	return node;
}

Ql_tree_node *
ql_tr_build_node_one_son(type, one)
	int type;
	Ql_tree_node *one;
{
	NODE(node, type, 1);
	node->kids[0] = one;
	return node;
}

Ql_tree_node *
ql_tr_build_node_two_sons(type, one, two)
	int	type;
	Ql_tree_node *one, *two;
{
	NODE(node, type, 2);
	node->kids[0] = one;
	node->kids[1] = two; 
	return node;
}

Ql_tree_node *
ql_tr_build_node_three_sons(type, one, two, three)
	int	type;
	Ql_tree_node *one, *two, *three;
{
	NODE(node, type, 3);
	node->kids[0] = one;
	node->kids[1] = two;
	node->kids[2] = three;
	return node;
}

Ql_tree_node *
ql_tr_build_node_four_sons(type, one, two, three, four)
	int	type;
	Ql_tree_node *one, *two, *three, *four;
{
	NODE(node, type, 4);
	node->kids[0] = one;
	node->kids[1] = two;
	node->kids[2] = three;
	node->kids[3] = four;
	return node;
}

Ql_tree_node *
ql_tr_add_left_son(me, son)
	Ql_tree_node *me, *son;
{
	Ql_tree_node **old, **new;
	int	i;

	old = me->kids;
	new = (Ql_tree_node **)ql_ma_synt_malloc((me->nkids+1) * sizeof(Ql_tree_node *));

	new[0] = son;
	for (i = 0; i < me->nkids; i++)
		new[i+1] = old[i];
	me->nkids++;
	me->kids = new;
	free(old);
	
	return me;
}

static indent(fp, level)
	FILE *fp;
	int level;
{
	int	i;
	for (i = 0; i < level ; i++)
		fputs("   ", fp);
}

extern char *ql_tr_name();

void
_ql_tr_print(fp, node, level)
	FILE *fp;
	Ql_tree_node *node;
	int	level;
{
	int	i;

	if (!node)
		return;
	/* indent(fp, level, NULL); */
	if (node->nkids == 0) {
		fprintf(fp, "%s, [id = %s]\n",
			ql_tr_name(node->type),  node->id);
		return;
	}
	fprintf(fp, "%s, %d kids:\n",
		ql_tr_name(node->type), node->nkids);

	for (i = 0; i < node->nkids; i++) {
		indent(fp, level+1);
		fprintf(fp, "[%d] ", i);
		_ql_tr_print(fp, node->kids[i], level+1);
	}
}

void
ql_tr_print(fp, node)
	FILE *fp;
	Ql_tree_node *node;
{
	_ql_tr_print(fp, node, 0);
}

#include "node_names.h"

char *ql_tr_name(int type)
{
	static char errbuf[128];
	if (type < 0 || type >= num_node_names) {
		sprintf(errbuf, "<unknown %d>", type);
		return errbuf;
	}
	return node_names[type];
}

void ql_tr_free(node)
	Ql_tree_node *node;
{
	int	i;

	/* Free my children */
	for (i = 0; i < node->nkids; i++) {
		ql_tr_free(node->kids[i]);
		node->kids[i] = (Ql_tree_node *)0;
	}
	/* Free the pointers to the kids */
	if (node->nkids)
		free(node->kids);
	/* dump the id */
	if (node->id)
		free(node->id);
	/* Free myself */
	free(node);
}
