#ifndef _SSTACK_H_
#define _SSTACK_H_


class SymbolStackItem {
	friend class SymbolStack;
private:
protected:
	SymbolStackItem	*_next;
public:
	SymbolTable	_table;
	SymbolStackItem(char *name) : _table(name) {
		_next = (SymbolStackItem *)0;
	}
};

class SymbolStack {
private:
	auto_string	_name;
	SymbolStackItem *_next;
public:
	SymbolStack(char *name = "<noname>") : _name(name) {
		_next = (SymbolStackItem *)0;
	}
	~SymbolStack() {
		if (_next)
			cerr << "Symbol stack " << _name.str()
				<< " not empty at destructor!" << endl;
	}

	
	void push(char *name) {
		SymbolStackItem *it = new SymbolStackItem(name);
		it->_next = _next;
		_next = it;
	}
#if 0
	SymbolTable *top() {
		return _next ? _next->_table : (SymbolTable *) 0;
	}
#endif
	void pop() {
		SymbolStackItem *it = _next;
		_next = it->_next;
		delete it;
	}
	bool	empty() { return _next == (SymbolStackItem *)0; }
	SymbolTable &top() {
		if (empty())
			cerr << "SymbolStack(" << _name.str() << ") empty!\n";
		return _next->_table;
	}
	Symbol *lookup_all(char *name) {
		SymbolStackItem *it = _next;
		Symbol *sym;
		while (it) {
			sym = it->_table.lookup(name);
			if (sym)
				return sym;
			it = it->_next;
		}
		return (Symbol *)0;
	}
	Symbol *lookup(char *name) {
		return top().lookup(name);
	}
	Symbol *enter(char *name, char *data = 0, int overwrite = 0) {
		return top().enter(name, data, overwrite);
	}
	Symbol	*overwrite(char *name, char *data = 0) {
		return top().overwrite(name, data);
	}
	int	remove(char *name) {
		return top().remove(name);
	}
};

#endif /* _SSTACK_H_ */
