
#include <ostream.h>
#include <strstream.h>
#include "symbol.hh"
#include "types.hh"

#include <assert.h>

PrimitiveType TypeVoid("void", PrimitiveType::pt_void);
ErrorType TypeError("*ERROR*");
ObjectType TypeNil("nil");

extern "C" {
	extern void bcopy(void *, void *, int);
}


ostream &operator<<(ostream &s, Type &type)
{
	return type.print(s);
}

#if 0
#include "primitive_names.h"
ostream &PrimitiveType::print(ostream &s)
{
	return s << '<' << PrimitiveNames[_primitive] << '>';
}
#endif


ostream &MethodType::print(ostream &s)
{
	int	i;
	Symbol	*sym;
	SymbolTable_i st(args);

	s << "<Method " << name() << '(';
	for (i = 0; sym = st.next(); ) {
		s << " ";
		s << sym->name() << ':' << *(Type *)(sym->data());
	}
	return s << " ) : " << _returns->name() << ">";
}


ostream &ContainerType::_print(ostream &s, char *typename)
{
	int	token = 0;
	Symbol	*sym;
	Type	*type;
	TableEntry	*entry;

	if (isNamed()) {
		/* kludge for objects */
		if (typename)
			return s << '<' << typename << '<' << _name << ">>";
		else
			return s << '<' << _name << '>';
	}

	s << '<' << typename << " <";

	SymbolTable_i	it(members);
	while (sym = it.next()) {
		s << sym->name() << ':';
		entry = (TableEntry *) (sym->data());
		type = entry->type;
		s << type->name();
		if (token >= 0)
			s << ' ';
	}
	s << ">>";
	return s;
}

ostream &CollectionType::_print(ostream &s, char *typename)
{
	return s << '<' << typename <<  *_of << ">";
}

Type	*RefType::rebuild(Type *of)
{
	return new RefType(of);
}

Type	*RefType::lcs(Type *_it)
{
	RefType *it = _it->isRef();
	Type	*of_lcs;

	if (!it)
		return 0;

	of_lcs = _of->lcs(it->_of);
	if (!of_lcs)
		return (Type *)0;

	return rebuild(of_lcs);
}

Type	*RefType::canCompareWith(Type *_it)
{
	return lcs(_it);
}

ostream &RefType::print(ostream &s)
{
	return s << "<Ref" << *_of << '>';
}

Type	*PrimitiveType::lcs(Type *_it)
{
	PrimitiveType *it = _it->isPrimitive();

	if (!it)
		return (Type *)0;

	if (_primitive == it->_primitive)
		return this;
	if ((_primitive == pt_real || _primitive == pt_integer) &&
	    (it->_primitive == pt_real  ||  it->_primitive == pt_integer)) {
		/* return an existing type, don't look it up in db! */
		if (_primitive == pt_real)
			return this;
		else
			return it;
	}
	return (Type *)0;
}

Type	*PrimitiveType::canCompareWith(Type *_it)
{
	return lcs(_it);
}

Type	*PrimitiveType::canDoMathWith(Type *_it)
{
	return lcs(_it);
}

Type	*PrimitiveType::canDoMath()
{
	return (_primitive == pt_real ||  _primitive == pt_integer) ?
		this : 0;
}

Type	*ContainerType::canCompareWith(Type *_it)
{
	return lcs(_it);
}

int	ObjectType::isa(ObjectType *it)
{
	if (it == this)
		return	1;
	if (!_superclass)
		return 0;
	return _superclass->isa(it);
}

void ObjectType::classlist(ObjectType **list, int &where, int number_left)
{
	assert(number_left > 0);
	list[where++] = this;
	if (_superclass)
		_superclass->classlist(list, where, number_left-1);
}

Type	*ObjectType::lcs(Type *_it)
{
	ObjectType *it;
	ObjectType **my_classes;
	ObjectType **its_classes;
	int	my_n, its_n;
	int	i, j;
	ObjectType *result;

	it = _it->isObject();
	if (!it)
		return (Type *)0;
	
	my_classes = new ObjectType *[100];
	my_n = 0;
	classlist(my_classes, my_n, 100);
	its_classes = new ObjectType *[100];
	its_n = 0;
	it->classlist(its_classes, its_n, 100);

	for (i = 0; i < my_n; i++)
		for (j = 0; j < its_n; j++)
			if (my_classes[i] == its_classes[j]) {
				result = my_classes[i];
				delete my_classes;
				delete its_classes;
				return result;
			}
	delete my_classes;
	delete its_classes;
	return (ObjectType *)0;
}


/* Add the tuple entries in the 'from' tuple to our tuple */
void	TupleType::concat(TupleType &from)
{
	Symbol	*sym;
	SymbolTable_i it(from.members);
	TableEntry	*entry;

	/* XXX assumes no collisions */
	while (sym = it.next()) {
		entry = (TableEntry *) (sym->data());
		add(sym->name(), entry->type);
	}
}

/*
   The LCS of two tuples, is the intersection of
   all attributes of the same name which have types which
   have LCSs
  */
Type	*TupleType::lcs(Type *_it)
{
	TupleType	*it = _it->isTuple();
	Type	*one,*two;
	int	i, j;
	Symbol	*is, *js;
	TupleType	*result = (TupleType *)0;
	Type	*lcs;

	if (!it)
		return (Type *)0;

	i = 0;
	SymbolTable_i outer(members);
	while (is = outer.next()) {
		j = 0;
		SymbolTable_i inner(it->members);
		while (js = inner.next()) {
			if (strcmp(is->name(), js->name()) != 0)
				continue;
			one = (Type *)(is->data());
			two = (Type *)(js->data());
#if 0
			lcs = ((Type *)(is->data()))->lcs(((Type *)(js->data())));
#else
			lcs = one->lcs(two);
#endif
			if (lcs) {
				cout << "Combine " << *one <<
					", " << *two <<'\n';
				cout << "\tInto " << *lcs << '\n'; 
				if (!result)
					result = new TupleType();
				result->add(is->name(), lcs);
			}
		    }
	}
	return result;
}

Type	*CollectionType::_lcs(CollectionType *it)
{
	Type	*of_lcs;

	/* CollectionType match already done by downwards lcs()s */

	of_lcs = _of->lcs(it->_of);
	if (!of_lcs)
		return (Type *)0;

	return rebuild(of_lcs);
}

Type	*BagType::lcs(Type *_it)
{
	BagType *it = _it->isBag();

	if (!it)
		return 0;

	return CollectionType::_lcs(it);
}

Type	*SetType::lcs(Type *_it)
{
	SetType *it = _it->isSet();

	/* allow set->bag downgrades, needed because of l->rebuild(r)  */
	if (!it && _it->isBag())
		return _it->lcs(this);

	return BagType::lcs(it);
}

Type	*ListType::lcs(Type *_it)
{
	ListType *it = _it->isList();

	if (!it)
		return 0;
	return CollectionType::_lcs(it);
}

Type	*ArrayType::lcs(Type *_it)
{
	ArrayType *it = _it->isArray();

	if (!it)
		return 0;
	return CollectionType::_lcs(it);
}


#if 1
char	*TupleType::name()
{
#if 0
//	return Type::name();
	char buf[1024];
	char buf2[128];
	Symbol *sym;
	int	token = 0;
	SymbolTable_i it(members);

	strcpy(buf, "<Tuple<");

	while (sym = it.next()) {
		sprintf(buf2, "%s:%s%s",
			sym->name(),
			((Type *)(sym->data()))->name(),
			token >= 0 ? " " : "");
		strcat(buf, buf2);
	}
	return strdup(strcat(buf, ">>"));	/* XXX leak */
#else
	ostrstream buf;
	buf << *this;
	return buf.str();	/* XXX leak */
#endif
}
#endif

char	*Type::name()
{
	return _name;
#if 0
	return strdup(form("#%#x", (int)this));
#else
//	cout << "Type::name() using tricks\n";
	ostrstream buf;
	buf << *this;
	return buf.str();	/* XXX leak */
#endif
}


int	PrimitiveType::size()
{
	int	s;

	switch (_primitive) {
	case pt_integer:
		s = sizeof(int);
		break;
	case pt_real:
		s = sizeof(double);
		break;
	case pt_boolean:
		s = sizeof(bool);
		break;
	case pt_char:
		s = sizeof(char);
		break;
	case pt_string:
		s = 0;
		break;
	case pt_void:
		s = 0;
		break;
	}

	return s;
}

int	PrimitiveType::inline_size()
{
	int	s;

	switch (_primitive) {
	case pt_string:
		s = sizeof(char *);
		break;
	default:
		s = size();
		break;
	}
	return s;
}

void	ContainerType::_commit()
{
	Symbol *sym;
	Type	*t;
	TableEntry	*entry;
	SymbolTable_i	it(members);

	_align = 1;	/* byte alignment */

	while (sym = it.next()) {
		entry = (TableEntry *) (sym->data());
		t = entry->type;
		t->commit();

		/* meta information; a method in this case */
		if (t->inline_size() == 0)
			continue;

		if ((_size % t->align()) != 0)
			_size += t->align() - (_size % t->align());
		entry->offset = _size;
#if 0
		cout.form("%s: '%s' (%d, %d) al=%d @ %d\n", 
			_name.str(), sym->name(), 
			t->size(), t->inline_size(), t->align(),
			entry->offset);
#endif
		_size += t->inline_size();

		/* alignment is most restrictive of all subtypes */
		if (t->align() > _align)
			_align = t->align();
	}
}

void	ContainerType::commit()
{
	/* commit already happened ? */
	if (commit_started)
		return;
	commit_started = true;
	_commit();
}

int	ContainerType::align()
{
	switch (_size) {
	case 1:
	case 2:
	case 4:
	case 8:
		return _size;
		break;
	case 3:
		return 4;
		break;
	case 5:
	case 6:
	case 7:
		return 8;
		break;
	default:
		return sizeof(double);
		break;
	}
}

/* XXX for multiple inheritance, we want to compute offsets
   w/in each chunk seperately

   an object then looks like a bunch of structs ina big struct

   aka
   object FUBAR : CRUD {
   	struct CRUD crud;
	struct FUBAR fubar;
   }

   alignment is confusing here
   is it the alignment of the total object,
   wor the alignment of this section ???

   Size is also confusing, since we may need to
   pad from the end of the parent;
 */

void	ObjectType::commit()
{
	/* this only allows for single inheritance in this implementation */
	if (_superclass)
		_superclass->commit();

	/* use normal commit to terminate recursion */
	ContainerType::commit();
}

#include "desc.hh"

ContainerType_i	*ContainerType::iterator()
{
	return new ContainerType_i(*this);
}

bool	ContainerType_i::next(char *&name, Type *&type, int &offset)
{
	Symbol	*sym;
	ContainerType::TableEntry	*entry;

	sym = _member.next();
	if (!sym)
		return false;

	entry = (ContainerType::TableEntry *) (sym->data());
	name = sym->name();
	type = entry->type;
	offset = entry->offset;

	return true;
}

MethodType_i	*MethodType::iterator()
{
	return new MethodType_i(*this);
}

bool	MethodType_i::next(char *&name, Type *&type)
{
	Symbol	*sym;

	sym = _args.next();
	if (!sym)
		return false;

	name = sym->name();
	type = (Type *) (sym->data());

	return true;
}


void	ContainerType::save(ostream &s)
{
}

void	MethodType::save(ostream &s)
{
	
}
